#include "nield.h"
#include "rtnetlink.h"

#define HIST_MAX 8

extern int optflag;

/* top of interface list and history */
static LIST_HEAD(lhead);
static LIST_HEAD(hhead);

/*
 * search an entry in an interface list
 */
static inline struct iflist_entry *search_iflist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * add an entry into an interface list
 */ 
static inline void add_iflist_entry(struct list_head *l)
{
	INIT_LIST_HEAD(l);
	list_add(l, &lhead);
}

/*
 * [not in use]
 * delete an entry in an interface list
 */ 
static inline void del_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_del(l);
			free(e);
		}
	}
}

/*
 * move an entry in an interface list
 */
static inline void move_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_move(l, &hhead);
			break;
		}
	}
}

/*
 * delete an entry in an interface history
 */
static inline void del_ifhist_entry(void)
{
	struct list_head *l, *n;
	int i = 0;

	list_for_each_safe(l, n, &hhead) {
		struct iflist_entry *e;

		if(++i > HIST_MAX) {
			e = list_entry(l, struct iflist_entry, list);
			list_del(l);
			free(e);
			break;
		}
	}
}

/*
 * search an entry in an interface history
 */ 
static inline struct iflist_entry *search_ifhist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &hhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * convert an interface index to name in an interface list
 */
char *if_indextoname_from_iflist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * convert an interface index to name in an inetrface history
 */ 
char *if_indextoname_from_ifhist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * get an interface type in an interface list
 */ 
unsigned short get_type_from_iflist(unsigned index)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * get an interface type in an interface history
 */
unsigned short get_type_from_ifhist(unsigned index)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * [not in use]
 * print an interface list
 */ 
void print_iflist(void)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);

		printf("%d ", e->index);
		printf("IFLA_IFNAME: %s, ", e->name);
		printf("IFLA_ADDRESS: %s, ", e->addr);
		printf("IFLA_BROADCAST: %s, ", e->brd);
		printf("IFLA_MTU: %hu\n", e->mtu);
	}
}

/*
 * create an interface list
 */ 
int create_iflist(struct msghdr *msg)
{
	struct nlmsghdr *nlh;
	int nlh_len;
	struct ifinfomsg *ifim;
	int ifim_len;
	struct rtattr *ifla[__IFLA_MAX];
	struct rtattr *linkinfo[__IFLA_INFO_MAX];
	struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
	unsigned char *addr;
	struct iflist_entry *ifle;

	/* get netlink message header */
	nlh = msg->msg_iov->iov_base;
	nlh_len = msg->msg_iov->iov_len;

	/* parse netlink message header */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		/* whether netlink message header ends or not */
		if(nlh->nlmsg_type == NLMSG_DONE)
			return(1);

		/* get interface infomation message */
		ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);
		ifim_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);

		/* parse interface infomation attributes */
		parse_ifinfo(ifla, nlh);

		/* debug */
		if(optflag & F_DEBUG) {
			debug_nlmsg(nlh);
			debug_ifimsg(ifim, ifla, ifim_len);
		}

		/* check bridge interface */
		if(ifim->ifi_family == PF_BRIDGE) {
			ifle = search_iflist(ifim->ifi_index);
			if(ifle)
				ifle->br_attached = 1;
		}

		/* check protocol family(PF_UNSPEC only) */
		if(ifim->ifi_family != PF_UNSPEC)
			continue;

		/* create interface list */
		ifle = (struct iflist_entry *)malloc(sizeof(struct iflist_entry));
		if(ifle == NULL) {
			rec_log("[%-12s] %s: malloc(): failed", "ERROR", __func__);
			return(-1);
		}
		memset(ifle, 0, sizeof(struct iflist_entry));
		INIT_LIST_HEAD(&(ifle->list));

		/* get interface name */
		ifle->index = ifim->ifi_index;
		if(ifla[IFLA_IFNAME])
			memcpy(ifle->name, RTA_DATA(ifla[IFLA_IFNAME]), sizeof(ifle->name));

		/* get interface address */
		if(ifla[IFLA_ADDRESS]) {
			char addrstr[INET6_ADDRSTRLEN+1] = "";
			if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
				RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
			memcpy(ifle->addr, addrstr, sizeof(addrstr));
		}

		/* get broadcast address */
		if(ifla[IFLA_BROADCAST]) {
			char brdstr[INET6_ADDRSTRLEN+1] = "";
			if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
				RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
			memcpy(ifle->brd, brdstr, sizeof(brdstr));
		}

		/* get interface broadcast address */
		if(ifla[IFLA_BROADCAST])
			memcpy(ifle->brd, RTA_DATA(ifla[IFLA_BROADCAST]), sizeof(ifle->brd));

		ifle->flags = ifim->ifi_flags;
		ifle->type = ifim->ifi_type;
		ifle->vid = 0;

		/* get interface information */
		if(ifla[IFLA_LINKINFO]) {
			parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
			if (linkinfo[IFLA_INFO_KIND]) {
				if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
					parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
					ifle->vid = *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]);
				} else {
					memcpy(ifle->kind, (char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]),
						sizeof(ifle->kind));
				}
			}
		}

		/* get interface MTU */
		if(ifla[IFLA_MTU])
			ifle->mtu = *(unsigned short *)RTA_DATA(ifla[IFLA_MTU]);

		/* get master interface */
		if(ifla[IFLA_MASTER])
			ifle->master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);

		/* add interface list */
		list_add(&(ifle->list), &lhead);
	}

	return(0);
}

/*
 * parse interface information messages
 */
int parse_ifimsg(struct nlmsghdr *nlh)
{
	struct ifinfomsg *ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);
	int ifim_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);
	struct rtattr *ifla[__IFLA_MAX];
	struct rtattr *linkinfo[__IFLA_INFO_MAX];
	struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
	struct iflist_entry *ifle_tmp, *ifle;
	char msg_type[16] = "", msg[128] = "";
	char *mp = msg;
	char ifname_master[IFNAMSIZ] = "";

	/* parse interface infomation message attributes */
	parse_ifinfo(ifla, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ifimsg(ifim, ifla, ifim_len);
	}

	/* set message type */
	strcpy(msg_type, "INTERFACE");

	/* create new interface list entry */
	ifle_tmp = malloc(sizeof(struct iflist_entry));
	if(!ifle_tmp) {
		rec_log("[%-12s] %s: malloc() failed", "ERROR", __func__);
		return(1);
	}
	memset(ifle_tmp, 0, sizeof(struct iflist_entry));
	INIT_LIST_HEAD(&(ifle_tmp->list));

	ifle_tmp->index = ifim->ifi_index;
	ifle_tmp->flags = ifim->ifi_flags;
	ifle_tmp->type = ifim->ifi_type;

	/* get interface name */
	if(ifla[IFLA_IFNAME]) {
		strncpy(ifle_tmp->name, RTA_DATA(ifla[IFLA_IFNAME]), IFNAMSIZ);
		mp += sprintf(mp, "if=%s ", ifle_tmp->name);
	}

	/* get interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		memcpy(ifle_tmp->addr, addrstr, sizeof(addrstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "local=%s ", ifle_tmp->addr);
				break;
			default:
				mp += sprintf(mp, "lladdr=%s ", ifle_tmp->addr);
		}
	}

	/* get broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		memcpy(ifle_tmp->brd, brdstr, sizeof(brdstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "remote=%s ", ifle_tmp->brd);
				break;
		}
	}

	/* get interface MTU */
	if(ifla[IFLA_MTU]) {
		ifle_tmp->mtu = *(unsigned *)RTA_DATA(ifla[IFLA_MTU]);
		mp += sprintf(mp, "mtu=%hu ", ifle_tmp->mtu);
	}

	/* get interface information */
	if(ifla[IFLA_LINKINFO]) {
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		if(linkinfo[IFLA_INFO_KIND]) {
			if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
				ifle_tmp->vid = *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]);
				mp += sprintf(mp, "vlan=%hu ", ifle_tmp->vid);
			} else {
				memcpy(ifle_tmp->kind, (char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]),
					sizeof(ifle_tmp->kind));
				mp += sprintf(mp, "info=%s ", ifle_tmp->kind);
			}
		}
	}

	/* get master interface */
	if(ifla[IFLA_MASTER]) {
		unsigned int ifindex_master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);
		ifle_tmp->master = ifindex_master;
		if(!if_indextoname(ifindex_master, ifname_master))
			if(!if_indextoname_from_iflist(ifindex_master, ifname_master))
				if_indextoname_from_ifhist(ifindex_master, ifname_master);
		mp += sprintf(mp, "master=%s ", ifname_master);
	}

	/* check protocol family(PF_BRIDGE only) & nlmsg_type */
	if(ifim->ifi_family == PF_BRIDGE) {
		if(nlh->nlmsg_type == RTM_NEWLINK) {
			ifle = search_iflist(ifle_tmp->index);
			if(ifle && !ifle->br_attached) {
				ifle->br_attached = 1;
				rec_log("[%-12s] interface %s attached to the ethernet bridge %s",
					msg_type, ifle->name, ifname_master);
			}
		} else if(nlh->nlmsg_type == RTM_DELLINK) {
			ifle = search_iflist(ifle_tmp->index);
			if(ifle && ifle->br_attached) {
				ifle->br_attached = 0;
				rec_log("[%-12s] interface %s detached from the ethernet bridge %s",
					msg_type, ifle->name, ifname_master);
			}
		} else {
			rec_log("[%-12s] unkonwn nlmsg_type: %d", "ERROR", nlh->nlmsg_type);
		}
	}

	/* check protocol family(PF_UNSPEC only) */
	if(ifim->ifi_family != PF_UNSPEC)
		return(0);

	/* check nlmsg_type */
	if(nlh->nlmsg_type == RTM_NEWLINK) {
		/* search interface list entry */
		ifle = search_iflist(ifle_tmp->index);
		if(ifle) {
			/* check bonding interface */
			if((ifle->flags & IFF_SLAVE) && !(ifle_tmp->flags & IFF_SLAVE)) {
				if(!if_indextoname_from_iflist(ifle->master, ifname_master))
					if_indextoname_from_ifhist(ifle->master, ifname_master);
				rec_log("[%-12s] interface %s detached from the bonding interface %s",
					msg_type, ifle_tmp->name, ifname_master);
			} else if(!(ifle->flags & IFF_SLAVE) && (ifle_tmp->flags & IFF_SLAVE))
				rec_log("[%-12s] interface %s attached to the bonding interface %s",
					msg_type, ifle_tmp->name, ifname_master);

			/* check administrative status */
			if((ifle->flags & IFF_UP) && !(ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to disabled", msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_UP) && (ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to enabled", msg_type, ifle_tmp->name);

			/* check operational status */
			if((ifle->flags & IFF_RUNNING) && !(ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to down", msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_RUNNING) && (ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to up", msg_type, ifle_tmp->name);

			ifle->flags = ifle_tmp->flags;

			/* check interface name */
			if(ifla[IFLA_IFNAME]) {
				if(strncmp(ifle->name, ifle_tmp->name, IFNAMSIZ)) {
					rec_log("[%-12s] interface name changed from %s to %s",
						msg_type, ifle->name, ifle_tmp->name);
					strncpy(ifle->name, ifle_tmp->name, IFNAMSIZ);
				}
			}

			/* check interface address */
			if(ifla[IFLA_ADDRESS]) {
				if(memcmp(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s local address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
							break;
						default:
							rec_log("[%-12s] interface %s link layer address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
					}
					memcpy(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr));
				}
			}

			/* check broadcast address */
			if(ifla[IFLA_BROADCAST]) {
				if(memcmp(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s remote address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->brd, ifle_tmp->brd);
							break;
					}
					memcpy(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd));
				}
			}

			ifle->type = ifle_tmp->type;

			/* check interface MTU */
			if(ifla[IFLA_MTU]) {
				if(ifle->mtu != ifle_tmp->mtu) {
					rec_log("[%-12s] interface %s mtu changed from %hu to %hu",
						msg_type, ifle->name, ifle->mtu, ifle_tmp->mtu);
					ifle->mtu = ifle_tmp->mtu;
				}
			} 

			/* check interface vlan id */
			if(ifle->vid != ifle_tmp->vid) {
				rec_log("[%-12s] interface %s vlan id changed from %hu to %hu",
					msg_type, ifle->name, ifle->vid, ifle_tmp->vid);
				ifle->vid = ifle_tmp->vid;
			}

			/* check master interface */
			if(ifle->master != ifle_tmp->master) {
				ifle->master = ifle_tmp->master;
			}

			free(ifle_tmp);
		} else {
			/* add interface list entry*/
			add_iflist_entry(&(ifle_tmp->list));

			/* check interface state */
			char state[20] = "";

			/* check administrative state */
			(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

			/* check operational state */
			(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

			rec_log("[%-12s] interface added: %sstate=%s", msg_type, msg, state);
		}
	} else if(nlh->nlmsg_type == RTM_DELLINK) {
		/* move entry from interface list to interface history */
		move_iflist_entry(ifle_tmp->index);
		del_ifhist_entry();

		/* delete interface information list */
		//del_iflist_entry(ifle_tmp->index);

		/* check interface state */
		char state[20] = "";

		/* check administrative state */
		(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

		/* check operational state */
		(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

		rec_log("[%-12s] interface deleted: %sstate=%s", msg_type, msg, state);

		free(ifle_tmp);
	} else {
		rec_log("[%-12s] unkonwn nlmsg_type: %d", "ERROR", nlh->nlmsg_type);
		free(ifle_tmp);
	}

	return(0);
}

/*
 * debug interface information messages
 */ 
void debug_ifimsg(struct ifinfomsg *ifim, struct rtattr *ifla[], int ifim_len)
{
	unsigned flags = ifim->ifi_flags;
	char flist[136] = "";
	int i;
	char *p;

#define _IFF(f) \
	if(flags & IFF_##f) \
		(flags &= ~IFF_##f) ? strcat(flist, #f ",") : strcat(flist, #f);
	_IFF(UP);
	_IFF(BROADCAST);
	_IFF(DEBUG);
	_IFF(LOOPBACK);
	_IFF(POINTOPOINT);
	_IFF(NOTRAILERS);
	_IFF(RUNNING);
	_IFF(NOARP);
	_IFF(PROMISC);
	_IFF(ALLMULTI);
	_IFF(MASTER);
	_IFF(SLAVE);
	_IFF(MULTICAST);
	_IFF(PORTSEL);
	_IFF(AUTOMEDIA);
	_IFF(DYNAMIC);
	_IFF(LOWER_UP);
	_IFF(DORMANT);
	_IFF(ECHO);
#undef _IFF

	/* debug ifinfomsg */
	rec_dbg("*********************************************************************");
	rec_dbg("[ ifinfomsg(%d) ]", NLMSG_ALIGN(sizeof(struct ifinfomsg)));
	rec_dbg("    ifi_family(%d): %d(%s)",
		sizeof(ifim->ifi_family), ifim->ifi_family, af_type_n2s(ifim->ifi_family));
	rec_dbg("    ifi_type(%d): %hu(%s)", sizeof(ifim->ifi_type), ifim->ifi_type, ifi_type_n2s(ifim->ifi_type));

	char ifname[IFNAMSIZ] = "";
	if(!if_indextoname(ifim->ifi_index, ifname))
		if(!if_indextoname_from_iflist(ifim->ifi_index, ifname))
			if_indextoname_from_ifhist(ifim->ifi_index, ifname);
	rec_dbg("    ifi_index(%d): %d(%s)", sizeof(ifim->ifi_index), ifim->ifi_index, ifname);
	rec_dbg("    ifi_flags(%d): 0x%.8x(%s)", sizeof(ifim->ifi_flags), ifim->ifi_flags, flist);
	rec_dbg("    ifi_change(%d): %u", sizeof(ifim->ifi_change), ifim->ifi_change);

	/* debug interface link attributes */
	rec_dbg("*********************************************************************");
	rec_dbg("[ ifinfomsg attributes(%d) ]", NLMSG_ALIGN(ifim_len - NLMSG_ALIGN(sizeof(struct ifinfomsg))));

	/* debug interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		rec_dbg("    IFLA_ADDRESS(%hu): %s", RTA_ALIGN(ifla[IFLA_ADDRESS]->rta_len), addrstr);
	}

	/* debug interface broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		rec_dbg("    IFLA_BROADCAST(%hu): %s", RTA_ALIGN(ifla[IFLA_BROADCAST]->rta_len), brdstr);
	}

	/* debug interface name */
	if(ifla[IFLA_IFNAME])
		rec_dbg("    IFLA_IFNAME(%hu): %s",
			RTA_ALIGN(ifla[IFLA_IFNAME]->rta_len), (char *)RTA_DATA(ifla[IFLA_IFNAME]));

	/* debug interface MTU */
	if(ifla[IFLA_MTU])
		rec_dbg("    IFLA_MTU(%hu): %hu",
			RTA_ALIGN(ifla[IFLA_MTU]->rta_len), *(unsigned short *)RTA_DATA(ifla[IFLA_MTU]));

	/* debug physical interface index */
	if(ifla[IFLA_LINK])
		rec_dbg("    IFLA_LINK(%hu): %u",
			RTA_ALIGN(ifla[IFLA_LINK]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_LINK]));

	/* debug interface queue discipline */
	if(ifla[IFLA_QDISC])
		rec_dbg("    IFLA_QDISC(%hu): %s",
			RTA_ALIGN(ifla[IFLA_QDISC]->rta_len), (char *)RTA_DATA(ifla[IFLA_QDISC]));

	/* debug interface statistics */
	if(ifla[IFLA_STATS]) {
		struct rtnl_link_stats *stats = (struct rtnl_link_stats *)RTA_DATA(ifla[IFLA_STATS]);
		rec_dbg("    IFLA_STATS(%hu):", RTA_ALIGN(ifla[IFLA_STATS]->rta_len));
		rec_dbg("        [ rtnl_link_stats(%d) ]", NLMSG_ALIGN(sizeof(struct rtnl_link_stats)));
		rec_dbg("            rx_packets(%d): %u", sizeof(stats->rx_packets), stats->rx_packets);
		rec_dbg("            tx_packets(%d): %u", sizeof(stats->tx_packets), stats->tx_packets);
		rec_dbg("            rx_bytes(%d): %u", sizeof(stats->rx_bytes), stats->rx_bytes);
		rec_dbg("            tx_bytes(%d): %u", sizeof(stats->tx_bytes), stats->tx_bytes);
		rec_dbg("            rx_errors(%d): %u", sizeof(stats->rx_errors), stats->rx_errors);
		rec_dbg("            tx_errors(%d): %u", sizeof(stats->tx_errors), stats->tx_errors);
		rec_dbg("            rx_dropped(%d): %u", sizeof(stats->rx_dropped), stats->rx_dropped);
		rec_dbg("            tx_dropped(%d): %u", sizeof(stats->tx_dropped), stats->tx_dropped);
		rec_dbg("            multicast(%d): %u", sizeof(stats->multicast), stats->multicast);
		rec_dbg("            collisions(%d): %u", sizeof(stats->collisions), stats->collisions);
		rec_dbg("            rx_length_errors(%d): %u",
			sizeof(stats->rx_length_errors), stats->rx_length_errors);
		rec_dbg("            rx_over_errors(%d): %u", sizeof(stats->rx_over_errors), stats->rx_over_errors);
		rec_dbg("            rx_crc_errors(%d): %u", sizeof(stats->rx_crc_errors), stats->rx_crc_errors);
		rec_dbg("            rx_frame_errors(%d): %u", sizeof(stats->rx_frame_errors), stats->rx_frame_errors);
		rec_dbg("            rx_fifo_errors(%d): %u", sizeof(stats->rx_fifo_errors), stats->rx_fifo_errors);
		rec_dbg("            rx_missed_errors(%d): %u",
			sizeof(stats->rx_missed_errors), stats->rx_missed_errors);
		rec_dbg("            tx_aborted_errors(%d): %u",
			sizeof(stats->tx_aborted_errors), stats->tx_aborted_errors);
		rec_dbg("            tx_carrier_errors(%d): %u",
			sizeof(stats->tx_carrier_errors), stats->tx_carrier_errors);
		rec_dbg("            tx_fifo_errors(%d): %u", sizeof(stats->tx_fifo_errors), stats->tx_fifo_errors);
		rec_dbg("            tx_heartbeat_errors(%d): %u",
			sizeof(stats->tx_heartbeat_errors), stats->tx_heartbeat_errors);
		rec_dbg("            tx_window_errors(%d): %u",
			sizeof(stats->tx_window_errors), stats->tx_window_errors);
		rec_dbg("            rx_compressed(%d): %u", sizeof(stats->rx_compressed), stats->rx_compressed);
		rec_dbg("            tx_compressed(%d): %u", sizeof(stats->tx_compressed), stats->tx_compressed);
	}

	/* unknown attribute */
	if(ifla[IFLA_COST])
		rec_dbg("    IFLA_COST(%hu): %u",
			RTA_ALIGN(ifla[IFLA_COST]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_COST]));

	/* unknown attribute */
	if(ifla[IFLA_PRIORITY])
		rec_dbg("    IFLA_PRIORITY(%hu): %u",
			RTA_ALIGN(ifla[IFLA_PRIORITY]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_PRIORITY]));

	/* debug master(parent) interface index */
	if(ifla[IFLA_MASTER]) {
		unsigned int ifindex_master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);
		char ifname_master[IFNAMSIZ] = "";
		if(!if_indextoname(ifindex_master, ifname_master))
			if(!if_indextoname_from_iflist(ifindex_master, ifname_master))
				if_indextoname_from_ifhist(ifindex_master, ifname_master);
		rec_dbg("    IFLA_MASTER(%hu): %u(%s)",
			RTA_ALIGN(ifla[IFLA_MASTER]->rta_len), ifindex_master, ifname_master);
	}

	/* debug wireless interface information(struct iw_event) */
	if(ifla[IFLA_WIRELESS])
		rec_dbg("    IFLA_WIRELESS(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_WIRELESS]->rta_len));

	/* debug interface protocol information */
	if(ifla[IFLA_PROTINFO])
		rec_dbg("    IFLA_PROTINFO(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_PROTINFO]->rta_len));

	/* debug interafce transmit queue length */
	if(ifla[IFLA_TXQLEN])
		rec_dbg("    IFLA_TXQLEN(%hu): %u",
			RTA_ALIGN(ifla[IFLA_TXQLEN]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_TXQLEN]));

	/* debug interface memory mapping */
	if(ifla[IFLA_MAP]) {
		struct rtnl_link_ifmap *ifmap = (struct rtnl_link_ifmap *)RTA_DATA(ifla[IFLA_MAP]);
		unsigned char *mem = (unsigned char *)&(ifmap->mem_start);
		rec_dbg("    IFLA_MAP(%hu):", RTA_ALIGN(ifla[IFLA_MAP]->rta_len));
		rec_dbg("        [ rtnl_link_ifmap(%d) ]", NLMSG_ALIGN(sizeof(struct rtnl_link_ifmap)));
		rec_dbg("            mem_start(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->mem_start), mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->mem_end);
		rec_dbg("            mem_end(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->mem_end), mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->base_addr);
		rec_dbg("            base_addr(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->base_addr), mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		rec_dbg("            irq(%d): %hu", sizeof(ifmap->irq), ifmap->irq);
		rec_dbg("            dma(%d): 0x%.2x", sizeof(ifmap->dma), ifmap->dma);
		rec_dbg("            port(%d): 0x%.2x", sizeof(ifmap->port), ifmap->port);
	}

	/* unknown attribute */
	if(ifla[IFLA_WEIGHT])
		rec_dbg("    IFLA_WEIGHT(%hu): %u",
			RTA_ALIGN(ifla[IFLA_WEIGHT]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_WEIGHT]));

	/* debug interface operational state(RFC 2863) */
	if(ifla[IFLA_OPERSTATE])
		rec_dbg("    IFLA_OPERSTATE(%hu): %d(%s)",
			RTA_ALIGN(ifla[IFLA_OPERSTATE]->rta_len),
			*(int *)RTA_DATA(ifla[IFLA_OPERSTATE]),
			if_oper_state_n2s(*(int *)RTA_DATA(ifla[IFLA_OPERSTATE])));

	/* debug interface link mode */
	if(ifla[IFLA_LINKMODE])
		rec_dbg("    IFLA_LINKMODE(%hu): %d(%s)",
			RTA_ALIGN(ifla[IFLA_LINKMODE]->rta_len),
			*(int *)RTA_DATA(ifla[IFLA_LINKMODE]),
			if_link_mode_n2s(*(int *)RTA_DATA(ifla[IFLA_LINKMODE])));

	/*
	 * debug interface link information
	 *
	 * IFLA_LINKINFO was committed in kernel 2.6.23.
	 *
	 * commit 38f7b870d4a6a5d3ec21557e849620cb7d032965
	 * Author: Patrick McHardy <kaber@trash.net>
	 * Date:   Wed Jun 13 12:03:51 2007 -0700
	 *
	 *     [RTNETLINK]: Link creation API
	 */
	if(ifla[IFLA_LINKINFO]) {
		struct rtattr *linkinfo[__IFLA_INFO_MAX];
		struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		rec_dbg("    IFLA_LINKINFO(%hu):", RTA_ALIGN(ifla[IFLA_LINKINFO]->rta_len));
		rec_dbg("        IFLA_INFO_KIND(%hu): %s",
			RTA_ALIGN(linkinfo[IFLA_INFO_KIND]->rta_len),
			(char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]));
		if(linkinfo[IFLA_INFO_KIND]) {
			if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
				rec_dbg("            IFLA_VLAN_ID(%hu): %d",
					RTA_ALIGN(vlaninfo[IFLA_INFO_KIND]->rta_len),
					*(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]));
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "bond")) {
				/* ignore */
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "tun")) {
				/* ignore */
			}
		}
	}

	/* 
	 * debug pid of process in network name space
	 *
	 * IFLA_NET_NS_PID was committed in kernel 2.6.24.
	 *
	 * commit d8a5ec672768c3cf4d51d7a63fc071520afa1617
	 * Author: Eric W. Biederman <ebiederm@xmission.com>
	 * Date:   Wed Sep 12 13:57:04 2007 +0200
	 *
	 *     [NET]: netlink support for moving devices between network namespaces.
	 */
	if(ifla[IFLA_NET_NS_PID])
		rec_dbg("    IFLA_NET_NS_PID(%hu): %u",
			RTA_ALIGN(ifla[IFLA_NET_NS_PID]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_NET_NS_PID]));

	/*
	 * debug interface alias name
	 *
	 * IFLA_IFALIAS was committed in kernel 2.6.28.
	 *
	 * commit 0b815a1a6d43ab498674b8430c8c35ab08487a16
	 * Author: Stephen Hemminger <shemminger@vyatta.com>
	 * Date:   Mon Sep 22 21:28:11 2008 -0700
	 *
	 *     net: network device name ifalias support
	 */ 
	if(ifla[IFLA_IFALIAS])
		rec_dbg("    IFLA_IFALIAS(%hu): %s",
			RTA_ALIGN(ifla[IFLA_IFALIAS]->rta_len), (char *)RTA_DATA(ifla[IFLA_IFALIAS]));

	/*
	 * debug number of VFs
	 *
	 * IFLA_NUM_VF was committed in kernel 2.6.34.
	 *
	 * commit b280da8d54b8d82b52f368a8703b7ada6c1744d5
	 * Author: Williams, Mitch A <mitch.a.williams@intel.com>
	 * Date:   Wed Feb 10 01:43:24 2010 +0000
	 *
	 *     if_link: Add SR-IOV configuration methods
	 */
	if(ifla[IFLA_NUM_VF])
		rec_dbg("    IFLA_NUM_VF(%hu): %u",
			RTA_ALIGN(ifla[IFLA_NUM_VF]->rta_len), *(unsigned *)RTA_DATA(ifla[IFLA_NUM_VF]));

	/*
	 * debug VFs information list
	 *
	 * IFLA_VFINFO_LIST was committed in kernel 2.6.34.
	 *
	 * commit c02db8c6290bb992442fec1407643c94cc414375
	 * Author: Chris Wright <chrisw@sous-sol.org>
	 * Date:   Sun May 16 01:05:45 2010 -0700
	 *
	 *     rtnetlink: make SR-IOV VF interface symmetric
	 */ 
	if(ifla[IFLA_VFINFO_LIST])
		rec_dbg("    IFLA_VFINFO_LIST(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_VFINFO_LIST]->rta_len));

	/*
	 * debug 64-bit interface statistics counter
	 *
	 * IFLA_STATS64 was committed in kernel 2.6.35.
	 *
	 * commit 10708f37ae729baba9b67bd134c3720709d4ae62
	 * Author: Jan Engelhardt <jengelh@medozas.de>
	 * Date:   Thu Mar 11 09:57:29 2010 +0000
	 *
	 *     net: core: add IFLA_STATS64 support
	 */
	if(ifla[IFLA_STATS64])
		rec_dbg("    IFLA_STATS64(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_STATS64]->rta_len));

	/*
	 * debug virtual port management
	 *
	 * IFLA_VF_PORTS and IFLA_PORT_SELF was committed in kernel 2.6.35.
	 *
	 * commit 57b610805ce92dbd79fc97509f80fa5391b99623
	 * Author: Scott Feldman <scofeldm@cisco.com>
	 * Date:   Mon May 17 22:49:55 2010 -0700
	 *
	 *     net: Add netlink support for virtual port management (was iovnl)
	 */
	if(ifla[IFLA_VF_PORTS])
		rec_dbg("    IFLA_VF_PORTS(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_VF_PORTS]->rta_len));

	if(ifla[IFLA_PORT_SELF])
		rec_dbg("    IFLA_PORT_SELF(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_PORT_SELF]->rta_len));

	/*
	 * debug link address family
	 *
	 * IFLA_AF_SPEC was committed in kernel 2.6.38.
	 *
	 * commit f8ff182c716c6f11ca3061961f5722f26a14e101
	 * Author: Thomas Graf <tgraf@infradead.org>
	 * Date:   Tue Nov 16 04:30:14 2010 +0000
	 *
	 *     rtnetlink: Link address family API
	 */
	if(ifla[IFLA_AF_SPEC])
		rec_dbg("    IFLA_AF_SPEC(%hu): -- ignored --", RTA_ALIGN(ifla[IFLA_AF_SPEC]->rta_len));

	rec_dbg("");

	return;
}

/*
 * convert an interafce type from number to string
 */ 
const char *ifi_type_n2s(int type)
{
	struct {
		int type;
		char *name;
	} ifi_tnames[] = {
#define _IFI_T(x) { x, #x }
		_IFI_T(ARPHRD_NETROM),
		_IFI_T(ARPHRD_ETHER),
		_IFI_T(ARPHRD_EETHER),
		_IFI_T(ARPHRD_AX25),
		_IFI_T(ARPHRD_PRONET),
		_IFI_T(ARPHRD_CHAOS),
		_IFI_T(ARPHRD_IEEE802),
		_IFI_T(ARPHRD_ARCNET),
		_IFI_T(ARPHRD_APPLETLK),
		_IFI_T(ARPHRD_DLCI),
		_IFI_T(ARPHRD_ATM),
		_IFI_T(ARPHRD_METRICOM),
		_IFI_T(ARPHRD_IEEE1394),
		_IFI_T(ARPHRD_EUI64),
		_IFI_T(ARPHRD_INFINIBAND),
		_IFI_T(ARPHRD_SLIP),
		_IFI_T(ARPHRD_CSLIP),
		_IFI_T(ARPHRD_SLIP6),
		_IFI_T(ARPHRD_CSLIP6),
		_IFI_T(ARPHRD_RSRVD),
		_IFI_T(ARPHRD_ADAPT),
		_IFI_T(ARPHRD_ROSE),
		_IFI_T(ARPHRD_X25),
		_IFI_T(ARPHRD_HWX25),
		_IFI_T(ARPHRD_CAN),
		_IFI_T(ARPHRD_PPP),
		_IFI_T(ARPHRD_CISCO),
		_IFI_T(ARPHRD_HDLC),
		_IFI_T(ARPHRD_LAPB),
		_IFI_T(ARPHRD_DDCMP),
		_IFI_T(ARPHRD_RAWHDLC),
		_IFI_T(ARPHRD_TUNNEL),
		_IFI_T(ARPHRD_TUNNEL6),
		_IFI_T(ARPHRD_FRAD),
		_IFI_T(ARPHRD_SKIP),
		_IFI_T(ARPHRD_LOOPBACK),
		_IFI_T(ARPHRD_LOCALTLK),
		_IFI_T(ARPHRD_FDDI),
		_IFI_T(ARPHRD_BIF),
		_IFI_T(ARPHRD_SIT),
		_IFI_T(ARPHRD_IPDDP),
		_IFI_T(ARPHRD_IPGRE),
		_IFI_T(ARPHRD_PIMREG),
		_IFI_T(ARPHRD_HIPPI),
		_IFI_T(ARPHRD_ASH),
		_IFI_T(ARPHRD_ECONET),
		_IFI_T(ARPHRD_IRDA),
		_IFI_T(ARPHRD_FCPP),
		_IFI_T(ARPHRD_FCAL),
		_IFI_T(ARPHRD_FCPL),
		_IFI_T(ARPHRD_FCFABRIC),
		_IFI_T(ARPHRD_IEEE802_TR),
		_IFI_T(ARPHRD_IEEE80211),
		_IFI_T(ARPHRD_IEEE80211_PRISM),
		_IFI_T(ARPHRD_IEEE80211_RADIOTAP),
		_IFI_T(ARPHRD_IEEE802154),
		_IFI_T(ARPHRD_PHONET),
		_IFI_T(ARPHRD_PHONET_PIPE)
#undef _IFI_T
	};
	int i;

	for(i = 0; i < sizeof(ifi_tnames)/sizeof(ifi_tnames[0]); i++)
		if(ifi_tnames[i].type == type)
			return ifi_tnames[i].name;

	return "unknown type";
}

/*
 * convert an interafce operational state from number to string
 */
const char *if_oper_state_n2s(int state)
{
	struct {
		int state;
		char *name;
	} if_osnames[] = {
#define _IF_OS(x) { x, #x }
		_IF_OS(IF_OPER_UNKNOWN),
		_IF_OS(IF_OPER_NOTPRESENT),
		_IF_OS(IF_OPER_DOWN),
		_IF_OS(IF_OPER_LOWERLAYERDOWN),
		_IF_OS(IF_OPER_TESTING),
		_IF_OS(IF_OPER_DORMANT),
		_IF_OS(IF_OPER_UP)
#undef _IF_OS
	};
	int i;

	for(i = 0; i < sizeof(if_osnames)/sizeof(if_osnames[0]); i++)
		if(if_osnames[i].state == state)
			return if_osnames[i].name;

	return "unknown state";
}

/*
 * convert an interafce link mode from number to string
 */
const char *if_link_mode_n2s(int mode)
{
	struct {
		int mode;
		char *name;
	} if_lmnames[] = {
#define _IF_LM(x) { x, #x }
		_IF_LM(IF_LINK_MODE_DEFAULT),
		_IF_LM(IF_LINK_MODE_DORMANT)
#undef _IF_LM
	};
	int i;

	for(i = 0; i < sizeof(if_lmnames)/sizeof(if_lmnames[0]); i++)
		if(if_lmnames[i].mode == mode)
			return if_lmnames[i].name;

	return "unknown mode";
}

/*
 * convert an interface address from number to string
 */ 
void if_addr_n2s(unsigned short type, unsigned char *addr, int addrlen, char *addrstr, int addrstrlen)
{
	int i;

	switch(type) {
		case ARPHRD_TUNNEL:
		case ARPHRD_IPGRE:
		case ARPHRD_SIT:
			inet_ntop(AF_INET, addr, addrstr, addrstrlen);
			break;
		case ARPHRD_TUNNEL6:
			inet_ntop(AF_INET6, addr, addrstr, addrstrlen);
			break;
		default:
			for(i = 0; i < addrlen; i++)
				addrstr += sprintf(addrstr, "%.2x:", addr[i]);
			addrstr--;
			*addrstr = '\0';
	}
}
